"""Views for the Fixed Income Extension."""

from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

from openbb_core.provider.abstract.data import Data

if TYPE_CHECKING:
    from openbb_charting.core.openbb_figure import (
        OpenBBFigure,
    )


class FixedIncomeViews:
    """FixedIncome Views."""

    @staticmethod
    def fixedincome_government_yield_curve(  # noqa: PLR0912
        **kwargs,
    ) -> Tuple["OpenBBFigure", Dict[str, Any]]:
        """Government Yield Curve Chart."""
        # pylint: disable=import-outside-toplevel
        from openbb_charting.charts.helpers import (
            duration_sorter,
        )
        from openbb_charting.core.chart_style import ChartStyle
        from openbb_charting.core.openbb_figure import OpenBBFigure
        from openbb_charting.styles.colors import LARGE_CYCLER
        from openbb_core.app.utils import basemodel_to_df
        from pandas import DataFrame

        data = kwargs.get("data", None)
        df: DataFrame = DataFrame()
        if data:
            if isinstance(data, DataFrame) and not data.empty:  # noqa: SIM108
                df = data
            elif isinstance(data, (list, Data)):
                df = basemodel_to_df(data, index=None)  # type: ignore
            else:
                pass
        else:
            df = DataFrame([d.model_dump() for d in kwargs["obbject_item"]])  # type: ignore

        if df.empty:
            raise ValueError("Error: No data to plot.")

        if "maturity" not in df.columns:
            raise ValueError("Error: Maturity column not found in the data.")

        if "rate" not in df.columns:
            raise ValueError("Error: Rate column not found in the data.")

        if "date" not in df.columns:
            raise ValueError("Error: Date column not found in the data.")

        provider = kwargs.get("provider")
        df["date"] = df["date"].astype(str)
        maturities = duration_sorter(df["maturity"].unique().tolist())

        # Use the supplied colors, if any.
        colors = kwargs.get("colors", [])
        if not colors:
            colors = LARGE_CYCLER
        color_count = 0

        figure = OpenBBFigure().create_subplots(shared_xaxes=True)
        figure.update_layout(ChartStyle().plotly_template.get("layout", {}))

        def create_fig(figure, df, dates, color_count, country: Optional[str] = None):
            """Create a scatter for each date in the data."""
            for date in dates:
                color = colors[color_count % len(colors)]
                plot_df = df[df["date"] == date].copy()
                plot_df["rate"] = plot_df["rate"].apply(lambda x: x * 100)
                plot_df = plot_df.rename(columns={"rate": "Yield"})
                plot_df = (
                    plot_df.drop(columns=["date"])
                    .set_index("maturity")
                    .filter(items=maturities, axis=0)
                    .reset_index()
                )
                plot_df = plot_df.rename(columns={"index": "Maturity"})
                plot_df["Maturity"] = [
                    (
                        d.split("_")[1] + " " + d.split("_")[0].title()
                        if d != "long_term"
                        else "Long Term"
                    )
                    for d in plot_df["Maturity"]
                ]
                figure.add_scatter(
                    x=plot_df["Maturity"],
                    y=plot_df["Yield"],
                    # fill=fill,
                    mode="lines+markers",
                    name=f"{country} - {date}" if country else date,
                    line=dict(width=3, color=color),
                    marker=dict(size=10, color=color),
                    hovertemplate=(
                        "Maturity: %{x}<br>Yield: %{y}%<extra></extra>"
                        if len(dates) == 1
                        else "%{fullData.name}<br>Maturity: %{x}<br>Yield: %{y}%<extra></extra>"
                    ),
                )
                color_count += 1
            return figure, color_count

        dates = df.date.unique().tolist()
        figure, color_count = create_fig(figure, df, dates, color_count)
        extra_params = kwargs.get("extra_params", {})
        extra_params = (
            extra_params if isinstance(extra_params, dict) else extra_params.__dict__
        )
        # Set the title for the chart
        country: str = ""
        if provider in ("federal_reserve", "fmp"):
            country = "United States"
        elif provider == "ecb":
            curve_type = (
                extra_params.get("yield_curve_type", "").replace("_", " ").title()
            )
            grade = extra_params.get("rating", "").replace("_", " ")
            grade = grade.upper() if grade == "aaa" else "All Ratings"
            country = f"Euro Area ({grade}) {curve_type}"
        elif provider == "fred":
            curve_type = extra_params.get("yield_curve_type", "")
            curve_type = (
                "Real Rates"
                if curve_type == "real"
                else curve_type.replace("_", " ").title()
            )
            country = f"United States {curve_type}"
        elif provider == "econdb":
            country = extra_params.get("country", "")
            country = country.replace("_", " ").title() if country else "United States"
        country = country + " " if country else ""
        title = kwargs.get("title", "")
        if not title:
            title = f"{country}Yield Curve"
            if len(dates) == 1:
                title = f"{country} Yield Curve - {dates[0]}"

        # Update the layout of the figure.
        figure.update_layout(
            title=dict(text=title, x=0.5, font=dict(size=20)),
            plot_bgcolor="rgba(255,255,255,0)",
            xaxis=dict(
                title="Maturity",
                ticklen=0,
                showgrid=False,
                type="category",
                categoryorder="array",
                categoryarray=(
                    [
                        (
                            d.split("_")[1] + " " + d.split("_")[0].title()
                            if d != "long_term"
                            else "Long Term"
                        )
                        for d in maturities
                    ]
                ),
            ),
            yaxis=dict(
                title="Yield (%)",
                ticklen=0,
                showgrid=True,
                gridcolor="rgba(128,128,128,0.3)",
            ),
            legend=dict(
                orientation="v",
                yanchor="top",
                xanchor="right",
                y=0.95,
                x=0,
                xref="paper",
                font=dict(size=12),
                bgcolor="rgba(0,0,0,0)",
            ),
            margin=dict(
                b=25,
                t=10,
            ),
        )

        layout_kwargs = kwargs.get("layout_kwargs", {})
        if layout_kwargs:
            figure.update_layout(layout_kwargs)

        content = figure.show(external=True)

        return figure, content



"""Fixed Income Corporate Router."""

from openbb_core.app.deprecation import OpenBBDeprecationWarning
from openbb_core.app.model.command_context import CommandContext
from openbb_core.app.model.example import APIEx
from openbb_core.app.model.obbject import OBBject
from openbb_core.app.provider_interface import (
    ExtraParams,
    ProviderChoices,
    StandardParams,
)
from openbb_core.app.query import Query
from openbb_core.app.router import Router

router = Router(prefix="/corporate")

# pylint: disable=unused-argument


@router.command(
    model="ICEBofA",
    examples=[
        APIEx(parameters={"provider": "fred"}),
        APIEx(parameters={"index_type": "yield_to_worst", "provider": "fred"}),
    ],
    deprecated=True,
    deprecation=OpenBBDeprecationWarning(
        message="This endpoint is deprecated; use `/fixedincome/bond_indices` instead.",
        since=(4, 2),
        expected_removal=(4, 5),
    ),
)
async def ice_bofa(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:
    """ICE BofA US Corporate Bond Indices.

    The ICE BofA US Corporate Index tracks the performance of US dollar denominated investment grade corporate debt
    publicly issued in the US domestic market. Qualifying securities must have an investment grade rating (based on an
    average of Moody’s, S&P and Fitch), at least 18 months to final maturity at the time of issuance, at least one year
    remaining term to final maturity as of the rebalance date, a fixed coupon schedule and a minimum amount
    outstanding of $250 million. The ICE BofA US Corporate Index is a component of the US Corporate Master Index.
    """
    return await OBBject.from_query(Query(**locals()))


@router.command(
    deprecated=True,
    deprecation=OpenBBDeprecationWarning(
        message="This endpoint is deprecated; use `/fixedincome/bond_indices` instead."
        + " Set `category` to `us` and `index` to `seasoned_corporate`.",
        since=(4, 2),
        expected_removal=(4, 5),
    ),
    model="MoodyCorporateBondIndex",
    examples=[
        APIEx(parameters={"provider": "fred"}),
        APIEx(parameters={"index_type": "baa", "provider": "fred"}),
    ],
)
async def moody(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:
    """Moody Corporate Bond Index.

    Moody's Aaa and Baa are investment bonds that acts as an index of
    the performance of all bonds given an Aaa or Baa rating by Moody's Investors Service respectively.
    These corporate bonds often are used in macroeconomics as an alternative to the federal ten-year
    Treasury Bill as an indicator of the interest rate.
    """
    return await OBBject.from_query(Query(**locals()))


@router.command(
    model="HighQualityMarketCorporateBond",
    examples=[
        APIEx(parameters={"provider": "fred"}),
        APIEx(parameters={"yield_curve": "par", "provider": "fred"}),
    ],
)
async def hqm(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:
    """High Quality Market Corporate Bond.

    The HQM yield curve represents the high quality corporate bond market, i.e.,
    corporate bonds rated AAA, AA, or A.  The HQM curve contains two regression terms.
    These terms are adjustment factors that blend AAA, AA, and A bonds into a single HQM yield curve
    that is the market-weighted average (MWA) quality of high quality bonds.
    """
    return await OBBject.from_query(Query(**locals()))


@router.command(
    model="SpotRate",
    examples=[
        APIEx(parameters={"provider": "fred"}),
        APIEx(parameters={"maturity": "10,20,30,50", "provider": "fred"}),
    ],
)
async def spot_rates(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:
    """Spot Rates.

    The spot rates for any maturity is the yield on a bond that provides a single payment at that maturity.
    This is a zero coupon bond.
    Because each spot rate pertains to a single cashflow, it is the relevant interest rate
    concept for discounting a pension liability at the same maturity.
    """
    return await OBBject.from_query(Query(**locals()))


@router.command(
    model="CommercialPaper",
    examples=[
        APIEx(parameters={"provider": "fred"}),
        APIEx(parameters={"category": "all", "maturity": "15d", "provider": "fred"}),
    ],
)
async def commercial_paper(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:
    """Commercial Paper.

    Commercial paper (CP) consists of short-term, promissory notes issued primarily by corporations.
    Maturities range up to 270 days but average about 30 days.
    Many companies use CP to raise cash needed for current transactions,
    and many find it to be a lower-cost alternative to bank loans.
    """
    return await OBBject.from_query(Query(**locals()))


@router.command(model="BondPrices", examples=[APIEx(parameters={"provider": "tmx"})])
async def bond_prices(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:
    """Corporate Bond Prices."""
    return await OBBject.from_query(Query(**locals()))



"""Fixed Income Router."""

# pylint: disable=W0613:unused-argument

from openbb_core.app.deprecation import OpenBBDeprecationWarning
from openbb_core.app.model.command_context import CommandContext
from openbb_core.app.model.example import APIEx
from openbb_core.app.model.obbject import OBBject
from openbb_core.app.provider_interface import (
    ExtraParams,
    ProviderChoices,
    StandardParams,
)
from openbb_core.app.query import Query
from openbb_core.app.router import Router

from openbb_fixedincome.corporate.corporate_router import router as corporate_router
from openbb_fixedincome.government.government_router import router as government_router
from openbb_fixedincome.rate.rate_router import router as rate_router
from openbb_fixedincome.spreads.spreads_router import router as spreads_router

router = Router(prefix="", description="Fixed Income market data.")
router.include_router(rate_router)
router.include_router(spreads_router)
router.include_router(government_router)
router.include_router(corporate_router)


@router.command(
    model="SOFR",
    examples=[
        APIEx(parameters={"provider": "fred"}),
    ],
    deprecated=True,
    deprecation=OpenBBDeprecationWarning(
        message="This endpoint is deprecated; use `/fixedincome/rate/sofr` instead.",
        since=(4, 2),
        expected_removal=(4, 5),
    ),
)
async def sofr(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:  # type: ignore
    """Secured Overnight Financing Rate.

    The Secured Overnight Financing Rate (SOFR) is a broad measure of the cost of
    borrowing cash overnight collateralizing by Treasury securities.
    """
    return await OBBject.from_query(Query(**locals()))


@router.command(
    model="BondIndices",
    examples=[
        APIEx(
            description="The default state for FRED are series for constructing the US Corporate Bond Yield Curve.",
            parameters={"provider": "fred"},
        ),
        APIEx(
            description="Multiple indices, from within the same 'category', can be requested.",
            parameters={
                "category": "high_yield",
                "index": "us,europe,emerging",
                "index_type": "total_return",
                "provider": "fred",
            },
        ),
        APIEx(
            description="From FRED, there are three main categories, 'high_yield', 'us', and 'emerging_markets'."
            + " Emerging markets is a broad category.",
            parameters={
                "category": "emerging_markets",
                "index": "corporate,private_sector,public_sector",
                "provider": "fred",
            },
        ),
    ],
)
async def bond_indices(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:  # type: ignore
    """Bond Indices."""
    return await OBBject.from_query(Query(**locals()))


@router.command(
    model="MortgageIndices",
    examples=[
        APIEx(
            description="The default state for FRED are the primary mortgage indices from Optimal Blue.",
            parameters={"provider": "fred"},
        ),
        APIEx(
            description="Multiple indices can be requested.",
            parameters={
                "index": "jumbo_30y,conforming_30y,conforming_15y",
                "provider": "fred",
            },
        ),
    ],
)
async def mortgage_indices(
    cc: CommandContext,
    provider_choices: ProviderChoices,
    standard_params: StandardParams,
    extra_params: ExtraParams,
) -> OBBject:  # type: ignore
    """Mortgage Indices."""
    return await OBBject.from_query(Query(**locals()))
